import ModernGL
import glfw
import glm

from meshcheck import events


class WindowError(RuntimeError):
    pass


class Window:
    def __init__(self, size, title, gl_version):
        if not glfw.init():
            raise WindowError('glfw.init failed')

        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, gl_version[0])
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, gl_version[1])
        self._wnd = glfw.create_window(*size, title, None, None)
        if not self._wnd:
            raise WindowError('glfw.create_window failed')
        self._init_events()

        self._ctx = None

    def _init_events(self):
        glfw.set_cursor_pos_callback(self._wnd, self._cursor_pos_callback)
        glfw.set_mouse_button_callback(self._wnd, self._mouse_button_callback)

    def _cursor_pos_callback(self, _, xpos, ypos):
        self.on_mouse_move(events.MouseMoveEvent(glm.vec2(xpos, ypos)))

    def _mouse_button_callback(self, _, button, action, mods):
        self.on_mouse_button(
            events.MouseButtonEvent(
                pos=self.get_mouse_position(),
                button=button,
                action=action,
                mods=mods))

    def get_mouse_position(self):
        return glm.vec2(*glfw.get_cursor_pos(self._wnd))

    def on_mouse_button(self, event):
        pass

    def on_mouse_move(self, event):
        pass

    def initialize(self, ctx):
        pass

    def render(self, ctx):
        pass

    def size(self):
        return glfw.get_framebuffer_size(self._wnd)

    def run(self):
        glfw.make_context_current(self._wnd)
        self._ctx = ModernGL.create_context()

        self.initialize(self._ctx)

        while not glfw.window_should_close(self._wnd):
            width, height = self.size()
            self._ctx.viewport = (0, 0, width, height)
            self.render(self._ctx)
            glfw.swap_buffers(self._wnd)
            glfw.wait_events()



from meshcheck import angle


class CameraController:
    """Simple mouse camera controller."""

    def __init__(self, camera):
        self._camera = camera
        self._orig_angle = None
        self._orig_elevation = None
        self._orig_pos = None
        self._in_drag = False

    @property
    def in_drag(self):
        return self._in_drag

    def start_drag(self, pos):
        """Start camera rotation."""
        self._orig_angle = self._camera.angle
        self._orig_elevation = self._camera.elevation
        self._orig_pos = pos
        self._in_drag = True

    def update_drag(self, pos):
        """Continue camera rotation."""
        delta = pos - self._orig_pos

        dampen = 2.0

        delta_angle = angle.Angle.from_degrees(delta.x / dampen)
        delta_elevation = angle.Angle.from_degrees(delta.y / dampen)

        self._camera.angle = self._orig_angle - delta_angle
        self._camera.elevation = self._orig_elevation + delta_elevation

    def end_drag(self):
        """End camera rotation."""
        self._in_drag = False



"""Shader-related utilities."""

import os

import attr


class ShaderError(RuntimeError):
    """Shader-related exception."""
    pass


# pylint can't find __slots__
#
# pylint: disable=no-member


@attr.s(frozen=True, slots=True)
class ShaderCode:
    """Shader code loader."""
    frag = attr.ib(default=None)
    geom = attr.ib(default=None)
    tess_ctrl = attr.ib(default=None)
    tess_eval = attr.ib(default=None)
    vert = attr.ib(default=None)

    @classmethod
    def load(cls, base_name):
        """Load a group of shaders from ./shaders/<base_name>.<kind>.glsl."""
        script_dir = os.path.dirname(os.path.realpath(__file__))
        shader_dir = os.path.join(script_dir, 'shaders')
        shader_dir = os.path.relpath(shader_dir)
        base_path = os.path.join(shader_dir, base_name)
        codes = cls(
            **{kind: cls._load_one(base_path, kind)
               for kind in cls.__slots__})
        if not codes.vert:
            raise ShaderError('vert shader is required')
        if not codes.vert:
            raise ShaderError('frag shader is required')
        return codes

    @staticmethod
    def _load_one(base_path, kind):
        try:
            path = '{}.{}.glsl'.format(base_path, kind)
            with open(path) as rfile:
                print('loading {}'.format(path))
                return rfile.read()
        except IOError:
            return None

    def create_shader(self, ctx, kind):
        """Create a ModernGL shader."""
        funcs = {
            'frag': 'fragment_shader',
            'geom': 'geometry_shader',
            'tess_ctrl': 'tess_control_shader',
            'tess_eval': 'tess_evaluation_shader',
            'vert': 'vertex_shader',
        }
        func = getattr(ctx, funcs[kind])
        code = getattr(self, kind)
        return func(code)

    def has_shader(self, kind):
        """Check if shader code of type |kind| is loaded."""
        return getattr(self, kind) is not None

    def create_program(self, ctx):
        """Create a ModernGL shader program."""
        shaders = [
            self.create_shader(ctx, kind) for kind in self.__slots__
            if self.has_shader(kind)
        ]
        return ctx.program(shaders)



import math

import cairo
import glm
import numpy

# pylint: disable=wrong-import-position
import gi
gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
from gi.repository import Pango, PangoCairo

from meshcheck import shader, util


def create_layout(cairo_ctx, text):
    layout = PangoCairo.create_layout(cairo_ctx)
    layout.set_font_description(Pango.FontDescription('serif 24'))
    auto_length = -1
    layout.set_text(text, auto_length)
    return layout


def measure_text(text):
    surface_size = (1, 1)
    surface = cairo.ImageSurface(cairo.FORMAT_A8, *surface_size)
    cairo_ctx = cairo.Context(surface)
    layout = create_layout(cairo_ctx, text)

    width, height = layout.get_size()
    scale = Pango.SCALE
    return (math.ceil(width / scale), math.ceil(height / scale))


def render(text):
    width, height = measure_text(text)
    # Add space for shadow
    width += 1
    height += 1
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    ctx = cairo.Context(surface)

    # Clear background
    ctx.set_operator(cairo.OPERATOR_SOURCE)
    ctx.set_source_rgba(0.0, 0.0, 0.0, 0.0)
    ctx.paint()

    ctx.set_source_rgba(0.0, 0.0, 0.0, 1.0)

    layout = create_layout(ctx, text)
    PangoCairo.show_layout(ctx, layout)

    ctx.set_source_rgba(1.0, 0.7, 0.8, 1.0)

    ctx.translate(-1, -1)
    PangoCairo.update_layout(ctx, layout)
    PangoCairo.show_layout(ctx, layout)

    return surface


class TextNode:
    Program = None

    def __init__(self, ctx, text, where):
        self._text = text
        self._where = where

        if TextNode.Program is None:
            shader_code = shader.ShaderCode.load('text')
            TextNode.Program = shader_code.create_program(ctx)

        self._vbo = self._make_vert_vbo(ctx)
        self._vao = self._make_tria_vao(ctx)
        self._texture = self._make_texture(ctx)

        TextNode.Program.uniforms['size'].value = self._texture.size

    @staticmethod
    def _make_vert_vbo(ctx):
        # Format: Two triangles with 2d position and 2d UV coord
        # yapf: disable
        arr = numpy.array([
            0, 0, 0, 1,
            1, 0, 1, 1,
            1, 1, 1, 0,
            0, 0, 0, 1,
            1, 1, 1, 0,
            0, 1, 0, 0
        ], dtype='float32')
        # yapf: enable
        return ctx.buffer(arr)

    def _make_tria_vao(self, ctx):
        attrs = ['pos', 'uv']
        return ctx.simple_vertex_array(self.Program, self._vbo, attrs)

    def _make_texture(self, ctx):
        surface = render(self._text)
        components = 4
        stride = surface.get_stride()
        width = surface.get_width()
        height = surface.get_height()
        # Remove the pixel data that comes from |stride|, ModernGL
        # doesn't have an input stride parameter currently
        src_data = surface.get_data()
        dst_data = bytes()
        for row in range(height):
            dst_data += src_data[row * stride:
                                 row * stride + width * components]
        return ctx.texture((width, height), components, dst_data)

    def render(self, proj, model_view):
        model_view = glm.translate(model_view, self._where)
        self.Program.uniforms['model_view'].write(util.to_gl(model_view))
        self.Program.uniforms['proj'].write(util.to_gl(proj))
        self._texture.use()
        self._vao.render()



import attr
import glfw


@attr.s
class MouseButtonEvent:
    pos = attr.ib()
    button = attr.ib()
    action = attr.ib()
    mods = attr.ib()

    @property
    def is_left_button(self):
        return self.button == glfw.MOUSE_BUTTON_LEFT

    @property
    def is_press(self):
        return self.action == glfw.PRESS

    @property
    def is_release(self):
        return self.action == glfw.RELEASE


@attr.s
class MouseMoveEvent:
    pos = attr.ib()



"""View/camera math."""

import math

import numpy
import glm

from meshcheck import angle


class Camera:
    """Simple angle-elevation camera."""

    def __init__(self):
        self._elevation = angle.Angle()
        self._angle = angle.Angle()
        self._distance = 5
        self._size = (1, 1)
        self._model_view = None
        self._proj = None
        self._mvp = None
        self._update()

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        self._angle = val
        self._update()

    @property
    def elevation(self):
        return self._elevation

    @elevation.setter
    def elevation(self, val):
        self._elevation = val
        self._update()

    @property
    def mvp(self):
        """Model-view-projection matrix."""
        return self._mvp

    @property
    def proj(self):
        """Projection matrix."""
        return self._proj

    @property
    def model_view(self):
        """Model-view matrix."""
        return self._model_view

    def aspect_ratio(self):
        height = self._size[1]
        if height == 0:
            return 1
        return self._size[0] / height

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        self._size = val
        self._update()

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, val):
        self._distance = val
        self._update()

    def _update(self):
        eye = numpy.array((self.angle.sin(), self.elevation.sin(),
                           self.angle.cos()))
        eye = eye * self.distance / numpy.linalg.norm(eye)

        target = (0, 0, 0)
        upv = (0, 1, 0)
        self._model_view = glm.lookAt(eye, target, upv)

        fovy = 70.0 * (math.pi / 180.0)
        aspect = self.aspect_ratio()
        near = 0.1
        far = 100.0
        self._proj = glm.perspective(fovy, aspect, near, far)

        self._mvp = self._proj * self._model_view



import math


class Angle:
    def __init__(self, radians=0):
        self._radians = radians

    @classmethod
    def from_radians(cls, radians):
        return cls(radians)

    @classmethod
    def from_degrees(cls, degrees):
        return cls(degrees * (math.pi / 180.0))

    @property
    def radians(self):
        return self._radians

    @property
    def degrees(self):
        return self.radians * (180.0 / math.pi)

    def cos(self):
        return math.cos(self.radians)

    def sin(self):
        return math.sin(self.radians)

    def tan(self):
        return math.tan(self.radians)

    def __add__(self, other):
        return Angle(self.radians + other.radians)

    def __sub__(self, other):
        return Angle(self.radians - other.radians)



import argparse
import json
import sys

import attr
import glm
import ModernGL
import numpy

from meshcheck import (camera, camera_controller, shader, text_render, util,
                       window)


class MeshCheckWindow(window.Window):
    def __init__(self, mesh):
        super().__init__((1280, 1024), 'meshcheck', (3, 3))
        self._vert_vbo = None
        self._tria_vao = None
        self._mesh = mesh
        self._vert_to_index = {}
        self._prog = None
        self._text_nodes = None
        self._mvp = None
        self._camera = camera.Camera()
        self._camera_controller = camera_controller.CameraController(
            self._camera)

    def _make_vert_vbo(self, ctx):
        lst = []
        for index, (key, value) in enumerate(self._mesh.verts.items()):
            self._vert_to_index[key] = index
            lst += value
        arr = numpy.array(lst, dtype='float32')
        return ctx.buffer(arr)

    def _make_tria_vbo(self, ctx):
        lst = []
        for tria in self._mesh.faces.values():
            # Naive triangle tessellation
            for index in range(2, len(tria)):
                for offset in (0, index - 1, index):
                    lst.append(self._vert_to_index[tria[offset]])

        arr = numpy.array(lst, dtype='uint32')
        return ctx.buffer(arr)

    def _make_tria_vao(self, ctx, prog):
        attrs = ['pos']
        fmt = ModernGL.buffers.detect_format(prog, attrs)
        vao = ctx.vertex_array(prog, [(self._vert_vbo, fmt, attrs)],
                               self._make_tria_vbo(ctx))
        return vao

    def on_mouse_button(self, event):
        if event.is_left_button:
            if event.is_press:
                self._camera_controller.start_drag(event.pos)
            elif event.is_release:
                self._camera_controller.end_drag()

    def on_mouse_move(self, event):
        if self._camera_controller.in_drag:
            self._camera_controller.update_drag(event.pos)

    def _face_center(self, face):
        center = glm.vec3()
        for vert in face:
            center += self._mesh.verts[vert]
        center /= len(face)
        return center

    def initialize(self, ctx):
        shader_code = shader.ShaderCode.load('default')
        self._prog = shader_code.create_program(ctx)
        self._mvp = self._prog.uniforms['mvp']

        self._text_nodes = [
            text_render.TextNode(ctx, key, vert)
            for key, vert in self._mesh.verts.items()
        ]

        self._text_nodes += [
            text_render.TextNode(ctx, key, self._face_center(face))
            for key, face in self._mesh.faces.items()
        ]

        self._vert_vbo = self._make_vert_vbo(ctx)
        self._tria_vao = self._make_tria_vao(ctx, self._prog)

    def render(self, ctx):
        ctx.clear(0.9, 0.9, 0.9)
        ctx.enable(ModernGL.DEPTH_TEST)
        ctx.enable(ModernGL.BLEND)
        ctx.disable(ModernGL.CULL_FACE)
        self._camera.size = self.size()
        self._mvp.write(util.to_gl(self._camera.mvp))
        self._tria_vao.render()
        for text_node in self._text_nodes:
            text_node.render(self._camera.proj, self._camera.model_view)


@attr.s
class Mesh:
    verts = attr.ib()
    faces = attr.ib()

    @classmethod
    def from_file(cls, path):
        with open(path) as rfile:
            return cls.from_json(json.load(rfile))

    @classmethod
    def from_stdin(cls):
        return cls.from_json(json.load(sys.stdin))

    @classmethod
    def from_json(cls, obj):
        return cls(verts=obj['verts'], faces=obj['faces'])


def parse_args():
    parser = argparse.ArgumentParser('simple 3D mesh visualizer')
    parser.add_argument('path', nargs='?')
    return parser.parse_args()


def run():
    args = parse_args()
    if args.path:
        mesh = Mesh.from_file(args.path)
    else:
        print('reading from stdin...')
        mesh = Mesh.from_stdin()
    wnd = MeshCheckWindow(mesh)
    wnd.run()
